from select_plus.src.models.models import CompressionTypes, CSVOutputSerialization, CSVInputSerialization, \
    JSONOutputSerialization, JSONInputSerialization, InputSerialization, OutputSerialization, EngineResults, EngineResultsStats
from tests.util.test_wrapper import TestWrapper


class TestModels(TestWrapper):

    def test_compression_types(self):
        c = CompressionTypes()
        self.assertEqual(c.none, 'NONE')
        self.assertEqual(c.bzip2, 'BZIP2')
        self.assertEqual(c.gzip, 'GZIP')

    def test_csv_input_serialization_valid(self):
        c = CSVInputSerialization(
            file_header_info='USE',
            comments='test1',
            quote_escape_character='test2',
            record_delimiter='test3',
            field_delimiter='test4',
            quote_character='test5',
            allow_quoted_record_delimiter=True
        )

        dict_repr = c.as_dict()
        expected = {'FileHeaderInfo': 'USE', 'Comments': 'test1', 'QuoteEscapeCharacter': 'test2',
                    'RecordDelimiter': 'test3', 'FieldDelimiter': 'test4', 'QuoteCharacter': 'test5',
                    'AllowQuotedRecordDelimiter': True}

        self.assertDictEqual(dict_repr, expected)

    def test_csv_input_serialization_valid_with_missing_params(self):
        c = CSVInputSerialization(
            file_header_info='USE',
            comments='test1',
            quote_escape_character='test2',
            record_delimiter='test3',
        )

        dict_repr = c.as_dict()
        expected = {'FileHeaderInfo': 'USE', 'Comments': 'test1', 'QuoteEscapeCharacter': 'test2',
                    'RecordDelimiter': 'test3'}

        self.assertDictEqual(dict_repr, expected)

    def test_csv_input_serialization_invalid(self):
        self.assertRaises(RuntimeError, CSVInputSerialization, file_header_info='wrong')

    def test_json_input_serialization_valid(self):
        c = JSONInputSerialization(
            Type='DOCUMENT'
        )

        dict_repr = c.as_dict()
        expected = {"Type": "DOCUMENT"}

        self.assertDictEqual(dict_repr, expected)

    def test_json_input_serialization_invalid(self):
        self.assertRaises(RuntimeError, JSONInputSerialization, Type='wrong')

    def test_input_serialization_with_compression_types(self):
        c = CompressionTypes()
        i = InputSerialization(
            compression_type=c.gzip,
            parquet={}
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'GZIP', 'Parquet': {}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_with_compression_str(self):
        i = InputSerialization(
            compression_type='GZIP',
            parquet={}
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'GZIP', 'Parquet': {}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_with_json_types(self):
        j = JSONInputSerialization(
            Type='DOCUMENT'
        )
        i = InputSerialization(
            json=j
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'NONE', 'JSON': {'Type': 'DOCUMENT'}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_with_json_str(self):
        i = InputSerialization(
            json={'Type': 'DOCUMENT'}
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'NONE', 'JSON': {'Type': 'DOCUMENT'}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_with_csv_types(self):
        c = CSVInputSerialization(
            file_header_info='USE'
        )
        i = InputSerialization(
            csv=c
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'NONE', 'CSV': {'FileHeaderInfo': 'USE'}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_with_csv_str(self):
        i = InputSerialization(
            csv={'FileHeaderInfo': 'USE'}
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'NONE', 'CSV': {'FileHeaderInfo': 'USE'}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_with_parquet_str(self):
        i = InputSerialization(
            parquet={}
        )

        dict_repr = i.as_dict()
        expected = {'CompressionType': 'NONE', 'Parquet': {}}
        self.assertDictEqual(dict_repr, expected)

    def test_input_serialization_fails_with_more_than_one_serializer(self):
        self.assertRaises(RuntimeError, InputSerialization, parquet={}, json={})

    def test_csv_output_serialization(self):
        c = CSVOutputSerialization(
            quote_fields='ALWAYS',
            quote_character='test1',
            quote_escape_character='test2',
            record_delimiter='test3',
            field_delimiter='test4'
        )

        dict_repr = c.as_dict()
        expected = {'QuoteFields': 'ALWAYS', 'QuoteEscapeCharacter': 'test2', 'RecordDelimiter': 'test3',
                    'FieldDelimiter': 'test4', 'QuoteCharacter': 'test1'}
        self.assertDictEqual(dict_repr, expected)

    def test_csv_output_serialization_raises(self):
        self.assertRaises(RuntimeError, CSVOutputSerialization, quote_fields='wrong')

    def test_json_output_serialization(self):
        j = JSONOutputSerialization(
            record_delimiter='test1',
        )

        dict_repr = j.as_dict()
        expected = {'RecordDelimiter': 'test1'}
        self.assertDictEqual(dict_repr, expected)

    def test_output_serialization_with_csv_types(self):
        c = CSVOutputSerialization()
        o = OutputSerialization(
            csv=c
        )

        dict_repr = o.as_dict()
        expected = {'CSV': {'QuoteFields': 'ASNEEDED'}}
        self.assertDictEqual(dict_repr, expected)

    def test_output_serialization_with_csv_str(self):
        o = OutputSerialization(
            csv={'QuoteFields': 'ASNEEDED'}
        )

        dict_repr = o.as_dict()
        expected = {'CSV': {'QuoteFields': 'ASNEEDED'}}
        self.assertDictEqual(dict_repr, expected)

    def test_output_serialization_with_json_types(self):
        j = JSONOutputSerialization()
        o = OutputSerialization(
            json=j
        )

        dict_repr = o.as_dict()
        expected = {'JSON': {}}
        self.assertDictEqual(dict_repr, expected)

    def test_output_serialization_with_json_str(self):
        o = OutputSerialization(
            json={'QuoteFields': 'ASNEEDED'}
        )

        dict_repr = o.as_dict()
        expected = {'JSON': {'QuoteFields': 'ASNEEDED'}}
        self.assertDictEqual(dict_repr, expected)

    def test_output_serialization_fails_with_more_than_one_serializer(self):
        self.assertRaises(RuntimeError, OutputSerialization, json={}, csv={})

    def test_engine_results_payload_csv(self):
        er = EngineResults(
            payload=['1,2,3,4,5,6,7,8,9,10\n1,2,3,4,5,6,7,8,9,10\n1,2,3,4,5,6,7,8,9,10\n1,2,3,4,5,6,7,8,9,10\n1,2,3,4,5,6,7,8,9,10\n'],
            stats=EngineResultsStats(
                cost=0,
                files_processed=0,
                bytes_scanned=0,
                bytes_returned=0,
                bytes_processed=0
            )
        )

        expected_result = [
            ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
            ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
            ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
            ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
            ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10']
        ]
        self.assertListEqual(expected_result, er.payload_csv)

    def test_engine_results_payload_dict(self):
        er = EngineResults(
            payload=['{"ticker":"VUSA.L"}\n{"ticker":"VUSA.L"}\n'],
            stats=EngineResultsStats(
                cost=0,
                files_processed=0,
                bytes_scanned=0,
                bytes_returned=0,
                bytes_processed=0
            )
        )

        expected_result = [{"ticker": "VUSA.L"}, {"ticker": "VUSA.L"}]
        self.assertListEqual(expected_result, er.payload_dict)

    def test_engine_results_payload_multiple_dict(self):
        er = EngineResults(
            payload=['[{"ticker":"VUSA.L"}\n],[{"ticker":"VUSA.L"}\n]'],
            stats=EngineResultsStats(
                cost=0,
                files_processed=0,
                bytes_scanned=0,
                bytes_returned=0,
                bytes_processed=0
            )
        )

        expected_result = [[{'ticker': 'VUSA.L'}], [{'ticker': 'VUSA.L'}]]
        self.assertListEqual(expected_result, er.payload_dict)


import unittest
import boto3
from moto import mock_s3
from select_plus.src.aws.s3 import S3
from select_plus.src.engine.base_engine import BaseEngine


class MockS3Paginator:

    @staticmethod
    def paginate(Bucket: str, Prefix: str):
        payload = [
            {
                'KeyCount': 1,
                'Contents': [{
                    "Key": 'test.json',
                    "Size": 123
                }]
            }
        ]

        return payload


class MockS3Client:

    @staticmethod
    def select_object_content(*args, **kwargs):
        payload = [
            {
                "Records": {
                    "Payload": str('test').encode()
                }
            },
            {
                "Stats": {
                    "Details": {
                        "BytesScanned": 1,
                        "BytesProcessed": 2,
                        "BytesReturned": 3
                    }
                }
            }
        ]

        mock_response = {"Payload": payload}

        return mock_response

    @staticmethod
    def get_paginator(*args, **kwargs):
        return MockS3Paginator()


class MockEngine(BaseEngine):
    def execute(self, sql_query: str, input_serialization: dict, output_serialization: dict,
                extra_func: callable = None, extra_func_args: dict = None, s3_client=None):
        response = [
            {"payload": "test",
             "stats": {
                "bytes_scanned": 30,
                "bytes_processed": 30,
                "bytes_returned": 10
                }
             }
        ]
        return response


@mock_s3
class TestWrapper(unittest.TestCase):

    def setUp(self) -> None:
        self.mock_s3_client = MockS3Client()
        self.client = boto3.client("s3", region_name='eu-west-1')

        # Create mock bucket
        self.client.create_bucket(Bucket='test-bucket',
                                  CreateBucketConfiguration={'LocationConstraint': 'eu-west-1'})
        self.s3 = S3(client=self.client)

        # Put a mock file in the bucket
        self.s3.client.put_object(
            Body=b'{"test": 1}',
            Bucket='test-bucket',
            Key='test-key/file.json'
        )

        # Create mock engine
        self.mock_engine = MockEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=False
        )





from abc import ABC, abstractmethod
from typing import Optional, Union
import boto3

from select_plus.src.aws.s3 import S3
from select_plus.src.models.models import InputSerialization, OutputSerialization


class BaseEngine(ABC):

    def __init__(self, bucket_name: str, prefix: str, threads: int, verbose: bool):
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.verbose = verbose
        self.threads = threads

    @abstractmethod
    def execute(self,
                sql_query: str,
                input_serialization: Union[InputSerialization, dict],
                output_serialization: Union[OutputSerialization, dict],
                extra_func: Optional[callable] = None,
                extra_func_args: Optional[dict] = None,
                s3_client: Optional[boto3.session.Session.client] = None
                ):
        raise NotImplementedError

    def select_s3(self,
                  key: str,
                  sql_query: str,
                  input_serialization: dict,
                  output_serialization: dict,
                  extra_func: callable,
                  extra_func_args: Optional[dict],
                  s3_client: Optional[boto3.session.Session.client] = None
                  ):
        s3 = S3(client=s3_client)
        response = s3.select(bucket_name=self.bucket_name, key=key, sql_string=sql_query,
                             input_serialization=input_serialization, output_serialization=output_serialization)
        if extra_func:
            response = self._apply_extra_func(response, extra_func, extra_func_args)

        return response

    @staticmethod
    def _apply_extra_func(response: dict, extra_func, extra_func_args):
        """
        A user has the possibility of adding an additional function at each thread level to process each chunk of data
        before it merges the results from all threads.
        Allow the function to access only the payload but not the statistics.
        This way, the cost can be computed in the compilation of the results after the proceses have ended.
        """

        block_response = {
            "stats": response['stats'],
            "payload": None
        }

        if extra_func_args:
            func_response = extra_func(response['payload'], **extra_func_args)
        else:
            func_response = extra_func(response['payload'])
        block_response['payload'] = func_response
        return block_response



from typing import Any
import boto3
from botocore.config import Config


class S3:

    def __init__(self, client: boto3.session.Session.client = None):
        self.config = Config(
            retries=dict(max_attempts=10)
        )
        self.client = client if client else boto3.client('s3', config=self.config)

    def put_object(self, bucket_name: str, key: str, body: Any):
        """
        Puts object into S3
        """
        self.client.put_object(Body=str(body).encode(), Bucket=bucket_name, Key=key)

    def list_objects(self, bucket_name: str, prefix: str):
        """
        Lists all objects in S3 with a prefix.
        """

        total_files = 0
        total_file_size = 0
        keys = []

        paginator = self.client.get_paginator('list_objects_v2')

        for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
            total_files += page['KeyCount']

            if total_files == 0:
                raise RuntimeError(f'No files found in prefix {prefix}')

            for content in page['Contents']:
                keys.append(content['Key'])
                total_file_size += content['Size']

        response = {
            "total_files": total_files,
            "total_file_size": total_file_size,
            "keys": keys
        }
        return response

    def select(self,
               bucket_name: str,
               key: str,
               sql_string: str,
               input_serialization: dict,
               output_serialization: dict
               ) -> dict:

        response = self.client.select_object_content(
            Bucket=bucket_name,
            Key=key,
            Expression=sql_string,
            ExpressionType='SQL',
            RequestProgress={
                'Enabled': True
            },
            InputSerialization=input_serialization,
            OutputSerialization=output_serialization
        )

        payload = list(response['Payload'])

        full_content = []
        bytes_scanned = None
        bytes_processed = None
        bytes_returned = None

        for item in payload:
            if "Records" in item.keys():
                full_content.append(item['Records']['Payload'].decode())
            if "Stats" in item.keys():
                bytes_scanned = item['Stats']['Details']['BytesScanned']
                bytes_processed = item['Stats']['Details']['BytesProcessed']
                bytes_returned = item['Stats']['Details']['BytesReturned']

        content = {
            "payload": ''.join(full_content),
            "stats": {
                "bytes_scanned": bytes_scanned,
                "bytes_processed": bytes_processed,
                "bytes_returned": bytes_returned,
            }
        }

        return content



from moto import mock_s3
from select_plus.ssp import SSP
from select_plus.src.engine.engine import EngineResults
from tests.util.test_wrapper import TestWrapper, MockEngine


@mock_s3
class TestSSP(TestWrapper):

    def test_estimate_cost(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        cost = ssp.estimate_cost()
        self.assertEqual(cost, 3.04e-11)

    def test_select_type(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )

        self.assertIsInstance(results, EngineResults)

    def test_select_results(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )
        expected_result = results.payload

        self.assertListEqual(expected_result, ['test'])

    def test_select_cost(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )
        expected_result = results.stats.cost

        self.assertEqual(expected_result, 6.77e-11)

    def test_select_files_processed(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )
        expected_result = results.stats.files_processed

        self.assertEqual(expected_result, 1)

    def test_select_bytes_processed(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )
        expected_result = results.stats.bytes_processed

        self.assertEqual(expected_result, 30)

    def test_select_bytes_scanned(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )
        expected_result = results.stats.bytes_scanned

        self.assertEqual(expected_result, 30)

    def test_select_bytes_returned(self):
        ssp = SSP(
            bucket_name='test-bucket',
            prefix='test-key',
            verbose=False,
            engine=MockEngine
        )

        results = ssp.select(
            sql_query='SELECT * FROM s3object s'
        )
        expected_result = results.stats.bytes_returned

        self.assertEqual(expected_result, 10)

    def test_raises_on_wrong_engine(self):
        class WrongEngine:
            pass

        self.assertRaises(RuntimeError, SSP, bucket_name='test-bucket', prefix='test-key', engine=WrongEngine)




from select_plus.src.models.models import InputSerialization, OutputSerialization, JSONOutputSerialization, \
    JSONInputSerialization, CSVOutputSerialization, CSVInputSerialization, CompressionTypes



from typing import Optional, Union

from select_plus.src.engine.base_engine import BaseEngine
from select_plus.src.utils.cost import Cost
from select_plus.src.models.models import EngineResults, EngineResultsStats, InputSerialization, OutputSerialization


class EngineWrapper:

    def execute(
            self,
            sql_query: str,
            extra_func: Optional[callable],
            extra_func_args: Optional[dict],
            engine: BaseEngine,
            input_serialization: Union[InputSerialization, dict],
            output_serialization: Union[OutputSerialization, dict]
    ) -> EngineResults:

        dict_input_serialization = self.deserialize(input_serialization)
        dict_output_serialization = self.deserialize(output_serialization)

        response = engine.execute(
            sql_query=sql_query,
            extra_func=extra_func,
            extra_func_args=extra_func_args,
            input_serialization=dict_input_serialization,
            output_serialization=dict_output_serialization
        )
        compiled_result = self._compile_results(response)

        return compiled_result

    @staticmethod
    def deserialize(obj) -> dict:
        """
        To allow users to have the serializers as either a defined class or a dictionary, this is required to convert
        the input / output serializers into dicts to be passed to the S3 select function
        """
        if isinstance(obj, OutputSerialization):
            return obj.as_dict()
        elif isinstance(obj, InputSerialization):
            return obj.as_dict()
        else:
            return obj

    @staticmethod
    def _compile_results(response: list) -> EngineResults:
        cost = Cost()

        payload = []
        bytes_scanned = 0
        bytes_processed = 0
        bytes_returned = 0
        files_processed = 0

        for record in response:
            payload.append(record['payload'])
            bytes_scanned += record['stats']['bytes_scanned']
            bytes_processed += record['stats']['bytes_processed']
            bytes_returned += record['stats']['bytes_returned']
            files_processed += 1

        cost = cost.compute_block(data_scanned=bytes_scanned,
                                  data_returned=bytes_returned,
                                  files_requested=files_processed)

        model = EngineResults(
            payload=payload,
            stats=EngineResultsStats(
                cost=cost,
                files_processed=files_processed,
                bytes_scanned=bytes_scanned,
                bytes_returned=bytes_returned,
                bytes_processed=bytes_processed
            )
        )

        return model



from select_plus.src.engine.engine import EngineWrapper, EngineResults
from tests.util.test_wrapper import TestWrapper


class TestEngineWrapper(TestWrapper):

    def test_execute_response_type(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertIsInstance(response, EngineResults)

    def test_response_payload_is_correct(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertListEqual(response.payload, ['test'])

    def test_response_cost_is_correct(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertEqual(response.stats.cost, 6.77e-11)

    def test_response_files_processed_is_correct(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertEqual(response.stats.files_processed, 1)

    def test_response_bytes_scanned_is_correct(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertEqual(response.stats.bytes_scanned, 30)

    def test_response_bytes_returned_is_correct(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertEqual(response.stats.bytes_returned, 10)

    def test_response_bytes_processed_is_correct(self):
        engine_wrapper = EngineWrapper()
        response = engine_wrapper.execute(sql_query='',
                                          extra_func=None,
                                          extra_func_args=None,
                                          engine=self.mock_engine,
                                          input_serialization={},
                                          output_serialization={})
        self.assertEqual(response.stats.bytes_processed, 30)


from typing import Type, Union
from multiprocessing import cpu_count

from select_plus.src.aws.s3 import S3
from select_plus.src.utils.cost import Cost
from select_plus.src.models.models import EngineResults, InputSerialization, OutputSerialization, \
    JSONInputSerialization, JSONOutputSerialization
from select_plus.src.engine.base_engine import BaseEngine
from select_plus.src.engine.sequential_engine import SequentialEngine
from select_plus.src.engine.parallel_engine import ParallelEngine
from select_plus.src.engine.engine import EngineWrapper


class SSP:

    def __init__(self,
                 bucket_name: str,
                 prefix: str = '',
                 engine: Union[Type[BaseEngine], ParallelEngine, SequentialEngine] = ParallelEngine,
                 verbose: bool = False):
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.verbose = verbose
        self.engine = engine

        # Validate the engine is inheriting the BaseEngine
        if not issubclass(engine, BaseEngine):
            raise RuntimeError(f'Engine {engine.__name__} does not inherit BaseEngine')

        self.cost = Cost()

    def estimate_cost(self) -> float:
        """
        Estimates the cost of selecting from all files
        """
        s3 = S3()
        s3_response = s3.list_objects(bucket_name=self.bucket_name, prefix=self.prefix)
        estimate_cost = self.cost.compute_block(
            data_scanned=s3_response['total_file_size'],
            data_returned=s3_response['total_file_size'],
            files_requested=s3_response['total_files']
        )
        return estimate_cost

    def select(
            self,
            sql_query: str,
            extra_func: callable = None,
            extra_func_args: dict = None,
            threads: int = cpu_count(),
            input_serialization: Union[InputSerialization, dict] = InputSerialization(
                json=JSONInputSerialization(Type='DOCUMENT')
            ),
            output_serialization: Union[OutputSerialization, dict] = OutputSerialization(
                json=JSONOutputSerialization()
            )
    ) -> EngineResults:
        eng = self.engine(bucket_name=self.bucket_name,
                          prefix=self.prefix,
                          threads=threads,
                          verbose=self.verbose)

        eng_wrapper = EngineWrapper()

        results = eng_wrapper.execute(sql_query=sql_query,
                                      extra_func=extra_func,
                                      extra_func_args=extra_func_args,
                                      engine=eng,
                                      input_serialization=input_serialization,
                                      output_serialization=output_serialization)

        return results





class Cost:

    def __init__(self):
        self.request_cost = 0.0004 / 1000
        self.returned_cost = 0.0007 / 1e+9  # dollars per byte
        self.scan_cost = 0.002 / 1e+9  # dollars per byte

    def compute_block(self, data_scanned: int, data_returned: int, files_requested: int) -> float:
        """
        Creates an estimate of the cost for a single file.
        """
        data_returned_cost = self.returned_cost * data_returned
        data_request_cost = self.returned_cost * files_requested
        data_scanned_cost = self.scan_cost * data_scanned

        total_cost = data_returned_cost + data_request_cost + data_scanned_cost

        return total_cost

    @staticmethod
    def compute_stack(cost_values: list) -> float:
        return float(f'{sum(cost_values):.9f}')



from typing import Optional
from multiprocessing import Pool
import tqdm
import boto3

from select_plus.src.engine.base_engine import BaseEngine
from select_plus.src.aws.s3 import S3


class ParallelEngine(BaseEngine):

    def execute(self,
                sql_query: str,
                input_serialization: dict,
                output_serialization: dict,
                extra_func: Optional[callable] = None,
                extra_func_args: Optional[dict] = None,
                s3_client: Optional[boto3.session.Session.client] = None,
                ) -> list:

        function_args = self._make_func_args(sql_query=sql_query, extra_func=extra_func,
                                             extra_func_args=extra_func_args, s3_client=s3_client,
                                             input_serialization=input_serialization,
                                             output_serialization=output_serialization)

        results = self.execute_callable(self._wrapper_func, function_args)
        return results

    def execute_callable(self, func: callable, args: list = None) -> list:
        """
        Generic parallel executor for a function with a list of arguments.
        The args must be of format [(arg1, arg2, arg3...), (arg1, arg2, arg3...)]
        """
        with Pool(self.threads) as pool:
            if self.verbose:
                print(f'Running with {self.threads} processes')
                result = list(tqdm.tqdm(pool.imap(func, args), total=len(args)))
            else:
                result = []
                partial_result = pool.imap(func, args)
                for i in partial_result:
                    result.append(i)
            return result

    def _make_func_args(self, sql_query: str, extra_func: callable, extra_func_args: dict, s3_client,
                        input_serialization, output_serialization) -> list:
        """
        Gets a list of all keys to be processed
        """
        s3 = S3(client=s3_client)
        s3_keys = s3.list_objects(bucket_name=self.bucket_name, prefix=self.prefix)
        keys = s3_keys['keys']
        func_args = []
        for key in keys:
            func_args.append({
                "key": key,
                "sql_query": sql_query,
                "extra_func": extra_func,
                "extra_func_args": extra_func_args,
                "s3_client": s3_client,
                "input_serialization": input_serialization,
                "output_serialization": output_serialization
            })

        return func_args

    def _wrapper_func(self, args):
        """
        This is a wrapper for the function to be executed inside the thread.
        The reason why this exists, is because if the tqdm bar is to exist, then tqdm doesn't work with "pool.starmap".
        As result, one cannot pass multiple parameters to the function.
        """
        result = self.select_s3(**args)
        return result



from moto import mock_s3
from tests.util.test_wrapper import TestWrapper
from select_plus.src.engine.sequential_engine import SequentialEngine


@mock_s3
class TestSequentialEngine(TestWrapper):

    def test_execute_with_verbose_flag(self):
        sequential_engine = SequentialEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=True
        )

        response = sequential_engine.execute(
            sql_query='',
            s3_client=self.mock_s3_client,
            input_serialization={},
            output_serialization={}
        )
        expected_response = [{'payload': 'test', 'stats': {'bytes_scanned': 1, 'bytes_processed': 2, 'bytes_returned': 3}}]
        self.assertListEqual(expected_response, response)

    def test_execute_without_verbose_flag(self):
        sequential_engine = SequentialEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=False
        )

        response = sequential_engine.execute(
            sql_query='',
            s3_client=self.mock_s3_client,
            input_serialization={},
            output_serialization={}
        )
        expected_response = [{'payload': 'test', 'stats': {'bytes_scanned': 1, 'bytes_processed': 2, 'bytes_returned': 3}}]
        self.assertListEqual(expected_response, response)

    def test_execute_with_extra_function(self):
        sequential_engine = SequentialEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=False
        )

        def extra_func(response, extra_field: str):
            response = f'{response}-{extra_field}'
            return response

        result = sequential_engine.execute(
            sql_query='',
            s3_client=self.mock_s3_client,
            extra_func=extra_func,
            extra_func_args={"extra_field": "test"},
            input_serialization={},
            output_serialization={}
        )

        expected_response = [
            {'stats': {'bytes_scanned': 1, 'bytes_processed': 2, 'bytes_returned': 3}, 'payload': 'test-test'}
        ]

        self.assertListEqual(expected_response, result)

    def test_execute_with_extra_function_no_args(self):
        sequential_engine = SequentialEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=False
        )

        def extra_func(response):
            return response

        result = sequential_engine.execute(
            sql_query='',
            s3_client=self.mock_s3_client,
            extra_func=extra_func,
            input_serialization={},
            output_serialization={}
        )

        expected_response = [
            {'stats': {'bytes_scanned': 1, 'bytes_processed': 2, 'bytes_returned': 3}, 'payload': 'test'}
        ]

        self.assertListEqual(expected_response, result)



from moto import mock_s3
from tests.util.test_wrapper import TestWrapper
from select_plus.src.engine.parallel_engine import ParallelEngine


# Do not put this function inside the test class. It cannot be pickled
def func(n: int):
    return n * n


@mock_s3
class TestParallelEngine(TestWrapper):

    def test_execute_callable_verbose(self):
        parallel_engine = ParallelEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=True
        )

        response = parallel_engine.execute_callable(
            func=func,
            args=[1, 2, 3, 4]
        )
        expected_response = [1, 4, 9, 16]
        self.assertListEqual(expected_response, response)

    def test_execute_callable(self):
        parallel_engine = ParallelEngine(
            bucket_name='test',
            prefix='test',
            threads=1,
            verbose=False
        )

        response = parallel_engine.execute_callable(
            func=func,
            args=[1, 2, 3, 4]
        )

        expected_response = [1, 4, 9, 16]
        self.assertListEqual(expected_response, response)

    def test_make_func_args(self):
        parallel_engine = ParallelEngine(
            bucket_name='test-bucket',
            prefix='test-key',
            threads=1,
            verbose=False
        )

        args_list = parallel_engine._make_func_args(
            sql_query='SELECT * FROM s3object s',
            extra_func=None,
            extra_func_args={'test': 1},
            s3_client=self.client,
            input_serialization={},
            output_serialization={}
        )

        args_list[0]['s3_client'] = None

        expected_args = [
            {
                'key': 'test-key/file.json',
                'sql_query': 'SELECT * FROM s3object s',
                'extra_func': None,
                's3_client': None,
                'extra_func_args': {'test': 1},
                'input_serialization': {},
                'output_serialization': {}
            }
        ]

        self.assertListEqual(args_list, expected_args)

    # Test cannot be executed because boto3 client cannot be pickled
    # def test_execute(self):
    #     parallel_engine = ParallelEngine(
    #         bucket_name='test-bucket',
    #         prefix='test-key',
    #         threads=1,
    #         verbose=False
    #     )
    #
    #     args_list = parallel_engine.execute(
    #         sql_query='SELECT * FROM s3object s',
    #         extra_func=None,
    #         extra_func_args={'test': 1},
    #         s3_client=self.client,
    #         input_serialization={},
    #         output_serialization={}
    #     )



from moto import mock_s3

from select_plus.src.aws.s3 import S3
from tests.util.test_wrapper import TestWrapper


@mock_s3
class TestS3(TestWrapper):

    def test_put_object(self):
        objects_in_bucket = self.s3.list_objects(bucket_name='test-bucket',
                                                 prefix='test')
        self.assertEqual(objects_in_bucket['total_files'], 1)

    def test_list_objects(self):
        objects_in_bucket = self.s3.list_objects(bucket_name='test-bucket',
                                                 prefix='test')
        keys = ['test-key/file.json']
        self.assertEqual(objects_in_bucket['keys'], keys)

    def test_list_objects_with_many_objects(self):
        for i in range(2000):
            self.s3.put_object(bucket_name='test-bucket',
                               key=f'test/test{i}.json',
                               body=b'{"test": 1}')

        objects_in_bucket = self.s3.list_objects(bucket_name='test-bucket',
                                                 prefix='test')

        self.assertEqual(len(objects_in_bucket['keys']), 2001)

    def test_list_objects_raises_without_files_present(self):
        self.assertRaises(RuntimeError, self.s3.list_objects, bucket_name='test-bucket', prefix='some-wrong-prefix')

    def test_list_objects_response(self):
        objects_in_bucket = self.s3.list_objects(bucket_name='test-bucket',
                                                 prefix='test')
        expected_response = {'keys': ['test-key/file.json'], 'total_file_size': 11, 'total_files': 1}
        self.assertEqual(objects_in_bucket, expected_response)

    # NO IMPLEMENTATION ON MOTO FOR S3 SELECT YET
    # A mocked class has been created to make sure the functionality is correct in the S3 class
    def test_select_json(self):
        s3 = S3(client=self.mock_s3_client)

        response = s3.select(
            bucket_name='test-bucket',
            key='test/test.json',
            sql_string='SELECT * FROM s3object[*] s',
            input_serialization={},
            output_serialization={}
        )

        expected_response = {'payload': 'test',
                             'stats': {'bytes_scanned': 1, 'bytes_processed': 2, 'bytes_returned': 3}}

        self.assertDictEqual(response, expected_response)



from select_plus.ssp import SSP, BaseEngine
from select_plus.src.engine.parallel_engine import ParallelEngine
from select_plus.src.engine.sequential_engine import SequentialEngine
from select_plus.src.models import models as serializer



from typing import Optional, Union
from dataclasses import dataclass


@dataclass
class EngineResultsStats:
    cost: float
    files_processed: int
    bytes_scanned: int
    bytes_returned: int
    bytes_processed: int


@dataclass
class EngineResults:
    payload: list
    stats: EngineResultsStats

    @property
    def payload_dict(self) -> list:
        payload_dict = []
        dict_repr = []
        for item in self.payload:
            sub_item = item.replace('\n', ',')
            record = list(eval(sub_item))
            dict_repr.append(record)

        for block in dict_repr:
            payload_dict += block

        return payload_dict

    @property
    def payload_csv(self):
        # CSV Parser
        payload_csv = []
        for item in self.payload:
            sub_item = item.split('\n')

            for row in sub_item:
                if len(row) > 0:
                    payload_csv.append(row.split(','))

        return payload_csv


@dataclass
class CompressionTypes:
    none: str = 'NONE'
    gzip: str = 'GZIP'
    bzip2: str = 'BZIP2'


@dataclass
class CSVInputSerialization:
    file_header_info: Optional[str] = None
    comments: Optional[str] = None
    quote_escape_character: Optional[str] = None
    record_delimiter: Optional[str] = None
    field_delimiter: Optional[str] = None
    quote_character: Optional[str] = None
    allow_quoted_record_delimiter: Optional[bool] = False

    def __post_init__(self):
        # Validate the file_header_info
        if self.file_header_info:
            if self.file_header_info not in ['USE', 'IGNORE', 'NONE']:
                raise RuntimeError('CSV Input Serialization field file_header_info must be USE | IGNORE | NONE')

    def as_dict(self) -> dict:
        params = {}
        if self.file_header_info:
            params['FileHeaderInfo'] = self.file_header_info
        if self.comments:
            params['Comments'] = self.comments
        if self.quote_escape_character:
            params['QuoteEscapeCharacter'] = self.quote_escape_character
        if self.record_delimiter:
            params['RecordDelimiter'] = self.record_delimiter
        if self.field_delimiter:
            params['FieldDelimiter'] = self.field_delimiter
        if self.quote_character:
            params['QuoteCharacter'] = self.quote_character
        if self.allow_quoted_record_delimiter:
            params['AllowQuotedRecordDelimiter'] = self.allow_quoted_record_delimiter

        return params


@dataclass
class JSONInputSerialization:
    Type: Optional[str] = None

    def __post_init__(self):
        # Validate the file_header_info
        if self.Type:
            if self.Type not in ['LINES', 'DOCUMENT']:
                raise RuntimeError('JSON Input Serialization field Type must be ALWAYS | DOCUMENT')

    def as_dict(self) -> dict:
        params = {}
        if self.Type:
            params['Type'] = self.Type
        return params


@dataclass
class InputSerialization:
    compression_type: Union[Optional[CompressionTypes], str] = CompressionTypes.none
    csv: Union[Optional[CSVInputSerialization], dict] = None
    json: Union[Optional[JSONInputSerialization], dict] = None
    parquet: dict = None

    def __post_init__(self):
        # At least one type must be set between CSV, JSON or PARQUET
        param_list = [self.csv, self.json, self.parquet]
        items_in_param_list = sum(x is not None for x in param_list)

        if items_in_param_list != 1:
            raise RuntimeError('InputSerializationError: Only one of the inputs must be selected: csv, json or parquet')

    def as_dict(self) -> dict:
        params = {}
        if self.compression_type:
            params['CompressionType'] = self.compression_type
        if self.csv:
            params['CSV'] = self.csv.as_dict() if isinstance(self.csv, CSVInputSerialization) else self.csv
        if self.json:
            params['JSON'] = self.json.as_dict() if isinstance(self.json, JSONInputSerialization) else self.json
        if self.parquet is not None:
            params['Parquet'] = {}
        return params


@dataclass
class CSVOutputSerialization:
    quote_fields: Optional[str] = 'ASNEEDED'
    quote_escape_character: Optional[str] = None
    record_delimiter: Optional[str] = None
    field_delimiter: Optional[str] = None
    quote_character: Optional[str] = None

    def __post_init__(self):
        # Validate the quote_fields
        if self.quote_fields:
            if self.quote_fields not in ['ALWAYS', 'ASNEEDED']:
                raise RuntimeError('CSV Output Serialization field quote_fields must be ALWAYS | ASNEEDED')

    def as_dict(self):
        params = {}
        if self.quote_fields:
            params['QuoteFields'] = self.quote_fields
        if self.quote_character:
            params['QuoteEscapeCharacter'] = self.quote_escape_character
        if self.record_delimiter:
            params['RecordDelimiter'] = self.record_delimiter
        if self.field_delimiter:
            params['FieldDelimiter'] = self.field_delimiter
        if self.quote_character:
            params['QuoteCharacter'] = self.quote_character

        return params


@dataclass
class JSONOutputSerialization:
    record_delimiter: Optional[str] = None

    def as_dict(self):
        params = {}
        if self.record_delimiter:
            params['RecordDelimiter'] = self.record_delimiter
        return params


@dataclass()
class OutputSerialization:
    csv: Union[Optional[CSVOutputSerialization], dict] = None
    json: Union[Optional[JSONOutputSerialization], dict] = None

    def __post_init__(self):
        # At least one type must be set between CSV, JSON
        param_list = [self.csv, self.json]
        items_in_param_list = sum(x is not None for x in param_list)

        if items_in_param_list != 1:
            raise RuntimeError('OutputSerializationError: Only one of the inputs must be selected: csv or json')

    def as_dict(self):
        params = {}
        if self.csv:
            params['CSV'] = self.csv.as_dict() if isinstance(self.csv, CSVOutputSerialization) else self.csv
        if self.json:
            params['JSON'] = self.json.as_dict() if isinstance(self.json, JSONOutputSerialization) else self.json
        return params



import unittest

from select_plus.src.utils.cost import Cost


class TestCost(unittest.TestCase):

    def test_compute_block(self):
        cost = Cost()
        result = cost.compute_block(
            data_scanned=100000,
            data_returned=10000,
            files_requested=10
        )
        self.assertEqual(result, 2.07007e-07)

    def test_compute_stack(self):
        cost = Cost()
        result = cost.compute_stack([1, 2, 3])
        self.assertEqual(result, 6)


import setuptools

setuptools.setup(
    name='s3select_plus',
    version='1.0.3',
    author="Gabriel Oana",
    author_email="gabriel.oana91@gmail.com",
    description="S3 select parallel package",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/gabriel-oana/s3select_plus',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'boto3>=1.24.75',
        'tqdm>=4.64.1'
    ]
)


from typing import Optional
import tqdm
import boto3
from select_plus.src.aws.s3 import S3
from select_plus.src.engine.base_engine import BaseEngine


class SequentialEngine(BaseEngine):

    def execute(self,
                sql_query: str,
                input_serialization: dict,
                output_serialization: dict,
                extra_func: Optional[callable] = None,
                extra_func_args: Optional[dict] = None,
                s3_client: Optional[boto3.session.Session.client] = None) -> list:
        s3 = S3(client=s3_client)
        s3_keys = s3.list_objects(bucket_name=self.bucket_name, prefix=self.prefix)
        keys = s3_keys['keys']

        result = []
        if self.verbose:
            itt = tqdm.tqdm(keys)
        else:
            itt = keys

        for key in itt:
            response = self.select_s3(key=key, sql_query=sql_query, extra_func=extra_func,
                                      extra_func_args=extra_func_args, s3_client=s3_client,
                                      input_serialization=input_serialization,
                                      output_serialization=output_serialization
                                      )
            result.append(response)

        return result



from unittest.mock import patch

from select_plus.src.engine.base_engine import BaseEngine
from tests.util.test_wrapper import TestWrapper


class TestBaseEngine(TestWrapper):

    def test_instance_raises(self):
        self.assertRaises(TypeError, BaseEngine, bucket_name='test', prefix='test', threads=1, verbose=False)

    @patch.multiple(BaseEngine, __abstractmethods__=set())
    def test_execute_raises(self):

        class TestEngine(BaseEngine):
            pass

        test_engine = TestEngine(bucket_name='test', prefix='test', threads=1, verbose=False)
        self.assertRaises(NotImplementedError, test_engine.execute, sql_query='', input_serialization={},
                          output_serialization={})
